# Copyright 2015 Futurewei. All rights reserved.
# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class SfcDriverBaseLegacy(metaclass=abc.ABCMeta):
    """SFC Driver Base Class for legacy interface."""

    @abc.abstractmethod
    def create_port_chain(self, context):
        pass

    @abc.abstractmethod
    def update_port_chain(self, context):
        pass

    @abc.abstractmethod
    def create_port_pair(self, context):
        pass

    @abc.abstractmethod
    def update_port_pair(self, context):
        pass

    @abc.abstractmethod
    def create_port_pair_group(self, context):
        pass

    @abc.abstractmethod
    def update_port_pair_group(self, context):
        pass


class SfcDriverBase(SfcDriverBaseLegacy, metaclass=abc.ABCMeta):
    """SFC Driver Base Class."""

    def create_port_chain_precommit(self, context):
        pass

    def create_port_chain_postcommit(self, context):
        self.create_port_chain(context)

    @abc.abstractmethod
    def delete_port_chain(self, context):
        pass

    def delete_port_chain_precommit(self, context):
        pass

    def delete_port_chain_postcommit(self, context):
        pass

    def update_port_chain_precommit(self, context):
        pass

    def update_port_chain_postcommit(self, context):
        self.update_port_chain(context)

    def create_port_pair_precommit(self, context):
        pass

    def create_port_pair_postcommit(self, context):
        self.create_port_pair(context)

    @abc.abstractmethod
    def delete_port_pair(self, context):
        pass

    def delete_port_pair_precommit(self, context):
        pass

    def delete_port_pair_postcommit(self, context):
        pass

    def update_port_pair_precommit(self, context):
        pass

    def update_port_pair_postcommit(self, context):
        self.update_port_pair(context)

    def create_port_pair_group_precommit(self, context):
        pass

    def create_port_pair_group_postcommit(self, context):
        self.create_port_pair_group(context)

    @abc.abstractmethod
    def delete_port_pair_group(self, context):
        pass

    def delete_port_pair_group_precommit(self, context):
        pass

    def delete_port_pair_group_postcommit(self, context):
        pass

    def update_port_pair_group_precommit(self, context):
        pass

    def update_port_pair_group_postcommit(self, context):
        self.update_port_pair_group(context)

    def create_service_graph_precommit(self, context):
        pass

    def create_service_graph_postcommit(self, context):
        pass

    def update_service_graph_precommit(self, context):
        pass

    def update_service_graph_postcommit(self, context):
        pass

    def delete_service_graph_precommit(self, context):
        pass

    def delete_service_graph_postcommit(self, context):
        pass



# Copyright (c) 2017 One Convergence Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from networking_sfc.extensions import sfc as sfc_ext
from networking_sfc.extensions import tap as tap_ext
from networking_sfc.tests.unit.extensions import test_sfc as test_sfc_ext
from oslo_utils import uuidutils
from webob import exc

from neutron.tests.unit.api.v2 import test_base as test_api_v2
import webtest


_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path


class SFCTapExtensionTestCase(test_sfc_ext.SfcExtensionTestCase):
    def setUp(self):
        super(test_sfc_ext.SfcExtensionTestCase, self).setUp()

        attr_map = sfc_ext.RESOURCE_ATTRIBUTE_MAP
        attr_map['port_pair_groups'].update(
            tap_ext.EXTENDED_ATTRIBUTES_2_0['port_pair_groups'])
        self.setup_extension(
            'networking_sfc.extensions.sfc.SfcPluginBase',
            sfc_ext.SFC_EXT,
            sfc_ext.Sfc,
            sfc_ext.SFC_PREFIX[1:],
            plural_mappings={}
        )

    def _get_expected_port_pair_group(self, data):
        ret = super(SFCTapExtensionTestCase,
                    self)._get_expected_port_pair_group(data)
        ret['port_pair_group'].update(
            tap_enabled=data['port_pair_group'].get('tap_enabled', False))
        return ret

    def test_create_port_pair_group_with_default_fields(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'tenant_id': tenant_id,
            'project_id': tenant_id,
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(
            _get_path(test_sfc_ext.PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port_pair_group', res)
        self.assertEqual(return_value, res['port_pair_group'])
        self.assertEqual(False, res['port_pair_group']['tap_enabled'])

    def test_create_port_pair_group_with_tap_enabled(self):
        portpairgroup_id = _uuid()
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'description': 'desc',
            'name': 'test1',
            'port_pairs': [],
            'port_pair_group_parameters': {},
            'tap_enabled': True,
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        expected_data = self._get_expected_port_pair_group(data)
        return_value = copy.copy(expected_data['port_pair_group'])
        return_value.update({'id': portpairgroup_id})
        instance = self.plugin.return_value
        instance.create_port_pair_group.return_value = return_value
        res = self.api.post(
            _get_path(test_sfc_ext.PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)
        instance.create_port_pair_group.assert_called_with(
            mock.ANY,
            port_pair_group=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertEqual(True, res['port_pair_group']['tap_enabled'])

    def test_create_port_pair_group_invalid_tap_enabled_value(self):
        tenant_id = _uuid()
        data = {'port_pair_group': {
            'port_pairs': [_uuid()],
            'tap_enabled': 'two',
            'tenant_id': tenant_id, 'project_id': tenant_id
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.post,
            _get_path(test_sfc_ext.PORT_PAIR_GROUP_PATH, fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)

    def test_update_port_pair_group_tap_enabled_field(self):
        portpairgroup_id = _uuid()
        data = {'port_pair_group': {
            'tap_enabled': True
        }}
        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(test_sfc_ext.PORT_PAIR_GROUP_PATH, id=portpairgroup_id,
                      fmt=self.fmt),
            self.serialize(data),
            content_type='application/%s' % self.fmt)



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defining flow-classifier data-model

Revision ID: 9768e6a66c9
Revises: 24fc7241aa5
Create Date: 2015-09-30 17:54:35.852573

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '9768e6a66c9'
down_revision = '24fc7241aa5'


def upgrade():
    op.create_table(
        'sfc_flow_classifiers',
        sa.Column('tenant_id', sa.String(length=255),
                  nullable=True, index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('ethertype', sa.String(length=40), nullable=True),
        sa.Column('protocol', sa.String(length=40), nullable=True),
        sa.Column('description', sa.String(length=255),
                  nullable=True),
        sa.Column('source_port_range_min', sa.Integer(), nullable=True),
        sa.Column('source_port_range_max', sa.Integer(), nullable=True),
        sa.Column('destination_port_range_min', sa.Integer(), nullable=True),
        sa.Column('destination_port_range_max', sa.Integer(), nullable=True),
        sa.Column('source_ip_prefix', sa.String(length=255), nullable=True),
        sa.Column('destination_ip_prefix', sa.String(length=255),
                  nullable=True),
        sa.Column('logical_source_port', sa.String(length=36),
                  nullable=False),
        sa.Column('logical_destination_port', sa.String(length=36),
                  nullable=True),
        sa.ForeignKeyConstraint(['logical_source_port'], ['ports.id'],
                                ondelete='RESTRICT'),
        sa.ForeignKeyConstraint(['logical_destination_port'], ['ports.id'],
                                ondelete='RESTRICT'),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sfc_flow_classifier_l7_parameters',
        sa.Column('keyword', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=True),
        sa.Column('classifier_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['classifier_id'],
                                ['sfc_flow_classifiers.id'], ),
        sa.PrimaryKeyConstraint('keyword', 'classifier_id')
    )



# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import logging
from unittest import mock

from neutron.api import extensions as api_ext
from neutron.common import config
import neutron.extensions as nextensions
from neutron_lib import constants as lib_const
from oslo_config import cfg
from oslo_utils import importutils
from oslo_utils import uuidutils
import webob.exc

from networking_sfc.db import flowclassifier_db as fdb
from networking_sfc import extensions
from networking_sfc.extensions import flowclassifier as fc_ext
from networking_sfc.tests import base


DB_FLOWCLASSIFIER_PLUGIN_CLASS = (
    "networking_sfc.db.flowclassifier_db.FlowClassifierDbPlugin"
)
extensions_path = ':'.join(extensions.__path__ + nextensions.__path__)


class FlowClassifierDbPluginTestCaseBase(base.BaseTestCase):
    def _create_flow_classifier(
        self, fmt, flow_classifier=None, expected_res_status=None, **kwargs
    ):
        ctx = kwargs.get('context', None)
        tenant_id = kwargs.get('tenant_id', self._tenant_id)
        data = {'flow_classifier': flow_classifier or {}}
        if ctx is None:
            data['flow_classifier'].update({'tenant_id': tenant_id})
        req = self.new_create_request(
            'flow_classifiers', data, fmt, context=ctx
        )
        res = req.get_response(self.ext_api)
        if expected_res_status:
            self.assertEqual(expected_res_status, res.status_int)
        return res

    @contextlib.contextmanager
    def flow_classifier(
        self, fmt=None, flow_classifier=None, do_delete=True, **kwargs
    ):
        if not fmt:
            fmt = self.fmt
        res = self._create_flow_classifier(fmt, flow_classifier, **kwargs)
        if res.status_int >= 400:
            logging.error('create flow classifier res: %s', res)
            raise webob.exc.HTTPClientError(
                code=res.status_int)
        flow_classifier = self.deserialize(fmt or self.fmt, res)
        yield flow_classifier
        if do_delete:
            self._delete('flow_classifiers',
                         flow_classifier['flow_classifier']['id'])

    def _get_expected_flow_classifier(self, flow_classifier):
        expected_flow_classifier = {
            'name': flow_classifier.get('name') or '',
            'description': flow_classifier.get('description') or '',
            'source_port_range_min': flow_classifier.get(
                'source_port_range_min'),
            'source_port_range_max': flow_classifier.get(
                'source_port_range_max'),
            'destination_port_range_min': flow_classifier.get(
                'destination_port_range_min'),
            'destination_port_range_max': flow_classifier.get(
                'destination_port_range_max'),
            'source_ip_prefix': flow_classifier.get(
                'source_ip_prefix'),
            'destination_ip_prefix': flow_classifier.get(
                'destination_ip_prefix'),
            'logical_source_port': flow_classifier.get(
                'logical_source_port'),
            'logical_destination_port': flow_classifier.get(
                'logical_destination_port'),
            'ethertype': flow_classifier.get(
                'ethertype') or 'IPv4',
            'protocol': flow_classifier.get(
                'protocol'),
            'l7_parameters': flow_classifier.get(
                'l7_parameters') or {}
        }
        return expected_flow_classifier

    def _assert_flow_classifiers_match_subsets(self, flow_classifiers,
                                               subsets, sort_key=None):
        # Sort both lists
        if sort_key:
            flow_classifiers.sort(key=lambda fc: fc[sort_key])
            subsets.sort(key=lambda fc: fc[sort_key])
        for fc, subset in zip(flow_classifiers, subsets):
            # Get matching items from subset
            sub_fc = dict([(k, fc[k]) for k
                           in subset.keys() if k in fc.keys()])
            self.assertEqual(subset, sub_fc)

    def _test_create_flow_classifier(
        self, flow_classifier, expected_flow_classifier=None
    ):
        if expected_flow_classifier is None:
            expected_flow_classifier = self._get_expected_flow_classifier(
                flow_classifier)
        with self.flow_classifier(flow_classifier=flow_classifier) as fc:
            for k, v in expected_flow_classifier.items():
                self.assertIn(k, fc['flow_classifier'])
                self.assertEqual(fc['flow_classifier'][k], v)


class FlowClassifierDbPluginTestCase(
    base.NeutronDbPluginV2TestCase,
    FlowClassifierDbPluginTestCaseBase
):
    resource_prefix_map = dict(
        (k, fc_ext.FLOW_CLASSIFIER_PREFIX)
        for k in fc_ext.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    def setUp(self, core_plugin=None, flowclassifier_plugin=None,
              ext_mgr=None):
        mock_log_p = mock.patch.object(fdb, 'LOG')
        self.mock_log = mock_log_p.start()
        cfg.CONF.register_opts(fc_ext.flow_classifier_quota_opts, 'QUOTAS')
        if not flowclassifier_plugin:
            flowclassifier_plugin = DB_FLOWCLASSIFIER_PLUGIN_CLASS
        service_plugins = {
            fc_ext.FLOW_CLASSIFIER_EXT: flowclassifier_plugin
        }
        fdb.FlowClassifierDbPlugin.supported_extension_aliases = [
            fc_ext.FLOW_CLASSIFIER_EXT]
        fdb.FlowClassifierDbPlugin.path_prefix = (
            fc_ext.FLOW_CLASSIFIER_PREFIX
        )
        super(FlowClassifierDbPluginTestCase, self).setUp(
            ext_mgr=ext_mgr,
            plugin=core_plugin,
            service_plugins=service_plugins
        )
        if not ext_mgr:
            self.flowclassifier_plugin = importutils.import_object(
                flowclassifier_plugin)
            ext_mgr = api_ext.PluginAwareExtensionManager(
                extensions_path,
                {fc_ext.FLOW_CLASSIFIER_EXT: self.flowclassifier_plugin}
            )
            app = config.load_paste_app('extensions_test_app')
            self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)

    def test_create_flow_classifier(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_logical_destination_port(self):
        with self.port(
            name='test1'
        ) as src_port, self.port(
            name='test1'
        ) as dst_port:
            self._test_create_flow_classifier({
                'logical_source_port': src_port['port']['id'],
                'logical_destination_port': dst_port['port']['id']
            })

    def test_quota_create_flow_classifier(self):
        cfg.CONF.set_override('quota_flow_classifier', 3, group='QUOTAS')
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.100.0.0/16',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=201)
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.101.0.0/16',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=201)
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.102.0.0/16',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=201)
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.103.0.0/16',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=409)

    def test_create_flow_classifier_with_all_fields(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'name': 'test1',
                'ethertype': lib_const.IPv4,
                'protocol': lib_const.PROTO_NAME_TCP,
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 101,
                'destination_port_range_max': 201,
                'source_ip_prefix': '10.100.0.0/16',
                'destination_ip_prefix': '10.200.0.0/16',
                'logical_source_port': port['port']['id'],
                'logical_destination_port': None,
                'l7_parameters': {}
            })

    def test_create_flow_classifier_with_all_supported_ethertype(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'ethertype': None,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'ethertype': 'IPv4',
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'ethertype': 'IPv6',
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_invalid_ethertype(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'ethertype': 'unsupported',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_all_supported_protocol(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'protocol': None,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'protocol': lib_const.PROTO_NAME_UDP,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'protocol': lib_const.PROTO_NAME_ICMP,
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_invalid_protocol(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'protocol': 'unsupported',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_all_supported_port_protocol(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'source_port_range_min': None,
                'source_port_range_max': None,
                'destination_port_range_min': None,
                'destination_port_range_max': None,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 100,
                'destination_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'source_port_range_min': 100,
                'source_port_range_max': 100,
                'destination_port_range_min': 100,
                'destination_port_range_max': 100,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'source_port_range_min': '100',
                'source_port_range_max': '200',
                'destination_port_range_min': '100',
                'destination_port_range_max': '200',
                'protocol': lib_const.PROTO_NAME_UDP,
                'logical_source_port': port['port']['id']
            }, {
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 100,
                'destination_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_UDP,
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_invalid_ip_prefix_ethertype(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '192.168.100.0/24',
                    'ethertype': 'IPv6',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': 'ff::0/24',
                    'ethertype': 'IPv4',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_invalid_port_protocol(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 'abc',
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_max': 'abc',
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 100,
                    'source_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 65536,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_max': 65536,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': -1,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_max': -1,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 'abc',
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_max': 'abc',
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 100,
                    'destination_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 65536,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_max': 65536,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': -1,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_max': -1,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 100,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_max': 100,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 100,
                    'source_port_range_max': 200,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_port_range_min': 100,
                    'source_port_range_max': 200,
                    'protocol': lib_const.PROTO_NAME_ICMP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 100,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_max': 100,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 100,
                    'destination_port_range_max': 200,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_port_range_min': 100,
                    'destination_port_range_max': 200,
                    'protocol': lib_const.PROTO_NAME_ICMP,
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_all_supported_ip_prefix(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'destination_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_invalid_ip_prefix(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.0.0.0/34',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.0.0.0.0/8',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '256.0.0.0/8',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '10.0.0.0',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_ip_prefix': '10.0.0.0/34',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_ip_prefix': '10.0.0.0.0/8',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_ip_prefix': '256.0.0.0/8',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'destination_ip_prefix': '10.0.0.0',
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_all_supported_l7_parameters(self):
        with self.port(
            name='test1'
        ) as port:
            self._test_create_flow_classifier({
                'l7_parameters': None,
                'logical_source_port': port['port']['id']
            })
            self._test_create_flow_classifier({
                'l7_parameters': {},
                'logical_source_port': port['port']['id']
            })

    def test_create_flow_classifier_with_invalid_l7_parameters(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'l7_parameters': {'abc': 'def'},
                    'logical_source_port': port['port']['id']
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_with_port_id(self):
        with self.port(
            name='test1'
        ) as port, self.port(
            name='test1'
        ) as port2:
            self._test_create_flow_classifier({
                'logical_source_port': port['port']['id'],
                'logical_destination_port': None,
            })
            self._test_create_flow_classifier({
                'logical_source_port': port['port']['id'],
                'logical_destination_port': port2['port']['id'],
            })

    def test_create_flow_classifier_with_nouuid_port_id(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'logical_source_port': 'abc'
                },
                expected_res_status=400
            )
            self._create_flow_classifier(
                self.fmt, {
                    'logical_source_port': port['port']['id'],
                    'logical_destination_port': 'abc'
                },
                expected_res_status=400
            )

    def test_create_flow_classifier_ethertype_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'ethertype': 'IPv4',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'ethertype': 'IPv6',
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'ethertype': 'IPv4',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'ethertype': 'IPv4',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'ethertype': 'IPv6',
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'ethertype': 'IPv6',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'ethertype': 'IPv6',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'ethertype': 'IPv4',
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_protocol_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'protocol': lib_const.PROTO_NAME_UDP,
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_source_ip_prefix_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'source_ip_prefix': '10.100.0.0/16'
                    },
                    expected_res_status=400
                )

            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '10.100.0.0/16',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'source_ip_prefix': '10.100.0.0/16',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_ip_prefix': '10.100.100.0/24',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_ip_prefix': '10.0.0.0/8',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'source_ip_prefix': '10.101.0.0/16',
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_destination_ip_prefix_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_ip_prefix': '10.100.0.0/16'
                    },
                    expected_res_status=400
                )

            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'destination_ip_prefix': '10.100.0.0/16',
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_ip_prefix': '10.100.0.0/16',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_ip_prefix': '10.100.100.0/24',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_ip_prefix': '10.0.0.0/8',
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'destination_ip_prefix': '10.101.0.0/16',
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_source_port_range_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 100,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 200,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 100,
                        'source_port_range_max': 200,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_port_range_min': 100,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'source_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_port_range_max': 100,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'source_port_range_min': 101,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 99,
                        'source_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 101,
                        'source_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 99,
                        'source_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 101,
                        'source_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_min': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'source_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'source_port_range_min': 201,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'source_port_range_min': 201,
                    'source_port_range_max': 300,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'source_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'source_port_range_min': 50,
                    'source_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_destination_port_range_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 100,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 200,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 100,
                        'destination_port_range_max': 200,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'destination_port_range_min': 100,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'destination_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'destination_port_range_max': 100,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'destination_port_range_min': 101,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'destination_port_range_min': 100,
                'destination_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._create_flow_classifier(
                    self.fmt, {
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 99,
                        'destination_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 101,
                        'destination_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 99,
                        'destination_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 101,
                        'destination_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 99,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_min': 101,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 199,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._create_flow_classifier(
                    self.fmt, {
                        'destination_port_range_max': 201,
                        'protocol': lib_const.PROTO_NAME_TCP,
                        'logical_source_port': port['port']['id']
                    },
                    expected_res_status=400
                )
                self._test_create_flow_classifier({
                    'destination_port_range_min': 201,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'destination_port_range_min': 201,
                    'destination_port_range_max': 300,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'destination_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
                self._test_create_flow_classifier({
                    'destination_port_range_min': 50,
                    'destination_port_range_max': 99,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_not_all_fields_conflict(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '192.168.100.0/24',
                'destination_ip_prefix': '192.168.101.0/24',
                'logical_source_port': port['port']['id']
            }):
                self._test_create_flow_classifier({
                    'source_ip_prefix': '192.168.100.0/24',
                    'destination_ip_prefix': '192.168.102.0/24',
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'destination_port_range_min': 100,
                'destination_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._test_create_flow_classifier({
                    'source_port_range_min': 100,
                    'source_port_range_max': 200,
                    'destination_port_range_min': 300,
                    'destination_port_range_max': 400,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '192.168.100.0/24',
                'source_port_range_min': 100,
                'source_port_range_max': 200,
                'protocol': lib_const.PROTO_NAME_TCP,
                'logical_source_port': port['port']['id']
            }):
                self._test_create_flow_classifier({
                    'source_ip_prefix': '192.168.100.0/24',
                    'source_port_range_min': 300,
                    'source_port_range_max': 400,
                    'protocol': lib_const.PROTO_NAME_TCP,
                    'logical_source_port': port['port']['id']
                })

    def test_create_flow_classifier_with_unknown_port_id(self):
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': uuidutils.generate_uuid()
                },
                expected_res_status=404
            )
            self._create_flow_classifier(
                self.fmt, {
                    'source_ip_prefix': '192.168.100.0/24',
                    'logical_source_port': port['port']['id'],
                    'logical_destination_port': uuidutils.generate_uuid()
                },
                expected_res_status=404
            )

    def test_list_flow_classifiers(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '10.100.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'name': 'test2',
                'source_ip_prefix': '10.101.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc2:
                fcs = [fc1, fc2]
                self._test_list_resources(
                    'flow_classifier', fcs
                )

    def test_list_flow_classifiers_with_params(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '10.100.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'name': 'test2',
                'source_ip_prefix': '10.101.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc2:
                self._test_list_resources(
                    'flow_classifier', [fc1],
                    query_params='name=test1'
                )
                self._test_list_resources(
                    'flow_classifier', [fc2],
                    query_params='name=test2'
                )
                self._test_list_resources(
                    'flow_classifier', [],
                    query_params='name=test3'
                )

    def test_list_flow_classifiers_with_unknown_params(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '10.100.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc1, self.flow_classifier(flow_classifier={
                'name': 'test2',
                'source_ip_prefix': '10.101.0.0/16',
                'logical_source_port': port['port']['id']
            }) as fc2:
                self._test_list_resources(
                    'flow_classifier', [fc1, fc2],
                    query_params='hello=test3'
                )

    def test_show_flow_classifier(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }) as fc:
                req = self.new_show_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                for k, v in fc['flow_classifier'].items():
                    self.assertEqual(res['flow_classifier'][k], v)

    def test_show_flow_classifier_noexist(self):
        req = self.new_show_request(
            'flow_classifiers', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)

    def test_update_flow_classifier(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'description': 'desc1',
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            }) as fc:
                updates = {
                    'name': 'test2',
                    'description': 'desc2',
                }
                req = self.new_update_request(
                    'flow_classifiers', {
                        'flow_classifier': updates
                    },
                    fc['flow_classifier']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                expected = fc['flow_classifier']
                expected.update(updates)
                for k, v in expected.items():
                    self.assertEqual(res['flow_classifier'][k], v)
                req = self.new_show_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = self.deserialize(
                    self.fmt, req.get_response(self.ext_api)
                )
                for k, v in expected.items():
                    self.assertEqual(res['flow_classifier'][k], v)

    def _test_update_with_field(
        self, fc, updates, expected_status_code
    ):
        req = self.new_update_request(
            'flow_classifiers', {'flow_classifier': updates},
            fc['flow_classifier']['id']
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(expected_status_code, res.status_int)

    def test_update_flow_classifer_unsupported_fields(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'description': 'desc1',
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            }) as fc:
                self._test_update_with_field(
                    fc, {'ethertype': None}, 400)
                self._test_update_with_field(
                    fc, {'protocol': None}, 400)
                self._test_update_with_field(
                    fc, {'source_port_range_min': None}, 400)
                self._test_update_with_field(
                    fc, {'source_port_range_max': None}, 400)
                self._test_update_with_field(
                    fc, {'destination_port_range_min': None}, 400)
                self._test_update_with_field(
                    fc, {'destination_port_range_max': None}, 400)
                self._test_update_with_field(
                    fc, {'source_ip_prefix': '192.168.100.0/24'}, 400)
                self._test_update_with_field(
                    fc, {'destination_ip_prefix': '192.168.100.0/24'}, 400)
                self._test_update_with_field(
                    fc, {'l7_parameters': None}, 400)

    def test_delete_flow_classifier(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'source_ip_prefix': '192.168.100.0/24',
                'logical_source_port': port['port']['id']
            }, do_delete=False) as fc:
                req = self.new_delete_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(204, res.status_int)
                req = self.new_show_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(404, res.status_int)

    def test_delete_flow_classifier_noexist(self):
        req = self.new_delete_request(
            'flow_classifiers', '1'
        )
        res = req.get_response(self.ext_api)
        self.assertEqual(404, res.status_int)



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.tests.functional import test_server

from networking_sfc.services.flowclassifier.common import (  # noqa
    config as fc_config)  # noqa
from networking_sfc.services.sfc.common import (  # noqa
    config as sfc_config)  # noqa


class TestService(test_server.TestPluginWorker):
    def test_start(self):
        cfg.CONF.set_override(
            'service_plugins', [
                'networking_sfc.services.flowclassifier.plugin.'
                'FlowClassifierPlugin',
                'networking_sfc.services.sfc.plugin.SfcPlugin'
            ]
        )
        cfg.CONF.set_override(
            'drivers', ['ovs'], group='flowclassifier'
        )
        cfg.CONF.set_override(
            'drivers', ['ovs'], group='sfc'
        )
        super(TestService, self).test_start()



# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Add Service Graph API resource

Revision ID: 53ed5bec6cff
Revises: 8329e9be2d8a
Create Date: 2017-05-24 00:00:00.000000

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '53ed5bec6cff'
down_revision = '8329e9be2d8a'


def upgrade():
    op.create_table('sfc_service_graphs',
                    sa.Column('project_id',
                              sa.String(length=255),
                              nullable=True),
                    sa.Column('id',
                              sa.String(length=36),
                              nullable=False),
                    sa.Column('name',
                              sa.String(length=255),
                              nullable=True),
                    sa.Column('description',
                              sa.String(length=255),
                              nullable=True),
                    sa.PrimaryKeyConstraint('id'),
                    mysql_engine='InnoDB')
    op.create_index(op.f('ix_sfc_service_graphs_project_id'),
                    'sfc_service_graphs',
                    ['project_id'],
                    unique=False)
    op.create_table('sfc_service_graph_chain_associations',
                    sa.Column('service_graph_id',
                              sa.String(length=36),
                              nullable=False),
                    sa.Column('src_chain',
                              sa.String(length=36),
                              nullable=False),
                    sa.Column('dst_chain',
                              sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['dst_chain'],
                                            ['sfc_port_chains.id'],
                                            ondelete='RESTRICT'),
                    sa.ForeignKeyConstraint(['service_graph_id'],
                                            ['sfc_service_graphs.id'],
                                            ondelete='CASCADE'),
                    sa.ForeignKeyConstraint(['src_chain'],
                                            ['sfc_port_chains.id'],
                                            ondelete='RESTRICT'),
                    sa.PrimaryKeyConstraint('service_graph_id',
                                            'src_chain',
                                            'dst_chain'),
                    mysql_engine='InnoDB')



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging

import oslo_messaging

from neutron_lib.agent import topics
from neutron_lib import rpc as n_rpc

from networking_sfc.services.sfc.drivers.ovs import rpc_topics as sfc_topics

LOG = logging.getLogger(__name__)


class SfcRpcCallback():
    """Sfc RPC server."""

    def __init__(self, driver):
        self.target = oslo_messaging.Target(version='1.0')
        self.driver = driver

    def get_flowrules_by_host_portid(self, context, **kwargs):
        host = kwargs.get('host')
        port_id = kwargs.get('port_id')
        LOG.debug('from port-chain service plugin')
        pcfrs = self.driver.get_flowrules_by_host_portid(
            context, host, port_id)
        LOG.debug('host: %s, port_id: %s', host, port_id)
        return pcfrs

    def update_flowrules_status(self, context, **kwargs):
        flowrules_status = kwargs.get('flowrules_status')
        LOG.info('update_flowrules_status: %s', flowrules_status)
        for flowrule_dict in flowrules_status:
            self.driver.update_flowrule_status(context,
                                               flowrule_dict['id'],
                                               flowrule_dict['status'])


class SfcAgentRpcClient():
    """RPC client for ovs sfc agent."""

    def __init__(self, topic=sfc_topics.SFC_AGENT):
        self.topic = topic
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def ask_agent_to_update_flow_rules(self, context, flow_rule):
        LOG.debug('Ask agent on the specific host to update flows ')
        LOG.debug('flow_rule: %s', flow_rule)
        host = flow_rule.get('host')
        cctxt = self.client.prepare(
            topic=topics.get_topic_name(
                self.topic, sfc_topics.PORTFLOW, topics.UPDATE),
            server=host)
        cctxt.cast(context, 'update_flow_rules', flowrule_entries=flow_rule)

    def ask_agent_to_delete_flow_rules(self, context, flow_rule):
        LOG.debug('Ask agent on the specific host to delete flows ')
        LOG.debug('flow_rule: %s', flow_rule)
        host = flow_rule.get('host')
        cctxt = self.client.prepare(
            topic=topics.get_topic_name(
                self.topic, sfc_topics.PORTFLOW, topics.DELETE),
            server=host)
        cctxt.cast(context, 'delete_flow_rules', flowrule_entries=flow_rule)



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FlowClassifierPluginContext():
    """Flow Classifier context base class."""
    def __init__(self, plugin, plugin_context):
        self._plugin = plugin
        self._plugin_context = plugin_context


class FlowClassifierContext(FlowClassifierPluginContext):

    def __init__(self, plugin, plugin_context, flowclassifier,
                 original_flowclassifier=None):
        super(FlowClassifierContext, self).__init__(plugin, plugin_context)
        self._flowclassifier = flowclassifier
        self._original_flowclassifier = original_flowclassifier

    @property
    def current(self):
        return self._flowclassifier

    @property
    def original(self):
        return self._original_flowclassifier



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_port_pair_group_params

Revision ID: fa75d46a7f11
Revises: d1002a1f97f6
Create Date: 2016-07-03 10:15:29.371910

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'fa75d46a7f11'
down_revision = 'd1002a1f97f6'


def upgrade():
    op.create_table('sfc_port_pair_group_params',
                    sa.Column('keyword', sa.String(length=255),
                              nullable=False),
                    sa.Column('value', sa.String(length=255),
                              nullable=True),
                    sa.Column('pair_group_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['pair_group_id'],
                                            ['sfc_port_pair_groups.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('keyword', 'pair_group_id'),
                    mysql_engine='InnoDB'
                    )
    op.add_column('sfc_port_chains', sa.Column('chain_id',
                  sa.Integer(), nullable=False))
    op.create_unique_constraint(None, 'sfc_port_chains', ['chain_id'])
    op.add_column('sfc_port_pair_groups', sa.Column('group_id',
                  sa.Integer(), nullable=False))
    op.create_unique_constraint(None, 'sfc_port_pair_groups', ['group_id'])



# Copyright 2016 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""update flow classifier

Revision ID: d1002a1f97f6
Revises: 5a475fc853e6
Create Date: 2016-06-03 10:23:52.850934

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd1002a1f97f6'
down_revision = '5a475fc853e6'


def upgrade():
    op.alter_column('sfc_flow_classifiers', 'logical_source_port',
                    nullable=True, existing_type=sa.String(length=36),
                    existing_nullable=False)



# Copyright 2017 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""_add_fwd_path_and_in_mac_column

Revision ID: b3adaf631bab
Revises: fa75d46a7f11
Create Date: 2016-10-27 17:01:16.793173

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'b3adaf631bab'
down_revision = 'fa75d46a7f11'


def upgrade():
    op.add_column('sfc_path_nodes', sa.Column('fwd_path', sa.Boolean(),
                                              nullable=False))
    op.add_column('sfc_portpair_details', sa.Column('in_mac_address',
                                                    sa.String(length=32),
                                                    nullable=True))



# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron.api.v2 import resource as api_res_log
from neutron import manager
from neutron.notifiers import nova as nova_log
from neutron.tests.unit.api.v2 import test_base as test_api_v2
from neutron.tests.unit.extensions import base as test_api_v2_extension
from oslo_config import cfg
from oslo_utils import uuidutils
from webob import exc
import webtest

from networking_sfc.extensions import servicegraph as sg_ext
from networking_sfc.extensions import sfc as sfc_ext

_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path

SERVICE_GRAPH_PATH = (sg_ext.SG_PREFIX[1:] + '/service_graphs')


class ServiceGraphExtensionTestCase(test_api_v2_extension.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        self._mock_unnecessary_logging()
        super(ServiceGraphExtensionTestCase, self).setUp()
        self.setup_extension(
            'networking_sfc.extensions.servicegraph.ServiceGraphPluginBase',
            sfc_ext.SFC_EXT,
            sg_ext.Servicegraph,
            sg_ext.SG_PREFIX[1:],
            plural_mappings={}
        )

    def _mock_unnecessary_logging(self):
        mock_log_cfg_p = mock.patch.object(cfg, 'LOG')
        self.mock_log_cfg = mock_log_cfg_p.start()

        mock_log_manager_p = mock.patch.object(manager, 'LOG')
        self.mock_log_manager = mock_log_manager_p.start()

        mock_log_nova_p = mock.patch.object(nova_log, 'LOG')
        self.mock_log_nova = mock_log_nova_p.start()

        mock_log_api_res_log_p = mock.patch.object(api_res_log, 'LOG')
        self.mock_log_api_res_log = mock_log_api_res_log_p.start()

    @staticmethod
    def _get_expected_service_graph(data):
        service_graph = data['service_graph']
        ret = {'service_graph': {
            'description': service_graph.get('description') or '',
            'name': service_graph.get('name') or '',
            'port_chains': service_graph['port_chains'],
            'tenant_id': service_graph['project_id'],
            'project_id': service_graph['project_id']
        }}
        return ret

    def _test_create_service_graph(self, **kwargs):
        tenant_id = _uuid()
        graph_data = {
            'port_chains': {_uuid(): [_uuid()]},
            'project_id': tenant_id
        }
        graph_data.update(kwargs)
        data = {'service_graph': graph_data}
        expected_data = self._get_expected_service_graph(data)
        return_value = copy.copy(expected_data['service_graph'])
        return_value.update({'id': _uuid()})
        instance = self.plugin.return_value
        instance.create_service_graph.return_value = return_value
        res = self.api.post(_get_path(SERVICE_GRAPH_PATH, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        instance.create_service_graph.assert_called_with(
            mock.ANY,
            service_graph=expected_data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('service_graph', res)
        self.assertEqual(return_value, res['service_graph'])

    def test_create_service_graph(self):
        self._test_create_service_graph()

    def test_create_service_graph_complex_dict(self):
        port_chains = {_uuid(): [_uuid()],
                       _uuid(): [_uuid(), _uuid()]}
        self._test_create_service_graph(description='desc',
                                        name='graph1',
                                        port_chains=port_chains)

    def test_list_service_graph(self):
        service_graph_id = _uuid()
        tenant_id = _uuid()
        return_value = [{
            'project_id': tenant_id,
            'id': service_graph_id
        }]
        instance = self.plugin.return_value
        instance.get_service_graphs.return_value = return_value

        res = self.api.get(_get_path(SERVICE_GRAPH_PATH, fmt=self.fmt))

        instance.get_service_graphs.assert_called_with(
            mock.ANY,
            fields=mock.ANY,
            filters=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('service_graphs', res)
        self.assertEqual(return_value, res['service_graphs'])

    def test_get_service_graph(self):
        service_graph_id = _uuid()
        tenant_id = _uuid()
        return_value = {
            'project_id': tenant_id,
            'id': service_graph_id
        }

        instance = self.plugin.return_value
        instance.get_service_graph.return_value = return_value

        res = self.api.get(_get_path(SERVICE_GRAPH_PATH,
                                     id=service_graph_id, fmt=self.fmt))

        instance.get_service_graph.assert_called_with(
            mock.ANY,
            service_graph_id,
            fields=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('service_graph', res)
        self.assertEqual(return_value, res['service_graph'])

    def test_update_service_graph(self):
        service_graph_id = _uuid()
        tenant_id = _uuid()
        update_data = {'service_graph': {
            'name': 'new_name',
            'description': 'new_desc'
        }}
        return_value = {
            'project_id': tenant_id,
            'id': service_graph_id
        }

        instance = self.plugin.return_value
        instance.update_service_graph.return_value = return_value

        res = self.api.put(_get_path(SERVICE_GRAPH_PATH, id=service_graph_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_service_graph.assert_called_with(
            mock.ANY, service_graph_id,
            service_graph=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('service_graph', res)
        self.assertEqual(return_value, res['service_graph'])

    def test_update_service_graph_with_port_chains(self):
        # API currently disallows rebuilding graphs, so we test this
        service_graph_id = _uuid()
        update_data = {'service_graph': {
            'name': 'new_name',
            'description': 'new_desc',
            'port_chains': {_uuid(): [_uuid()]}
        }}

        self.assertRaises(
            webtest.app.AppError,
            self.api.put,
            _get_path(SERVICE_GRAPH_PATH, id=service_graph_id, fmt=self.fmt),
            self.serialize(update_data),
            content_type='application/%s' % self.fmt)

    def test_delete_service_graph(self):
        self._test_entity_delete('service_graph')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_policy import policy

from networking_sfc.policies import base


rules = [
    policy.DocumentedRuleDefault(
        'create_flow_classifier',
        base.RULE_ANY,
        'Create a flow classifier',
        [
            {
                'method': 'POST',
                'path': '/sfc/flow_classifiers',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'update_flow_classifier',
        base.RULE_ADMIN_OR_OWNER,
        'Update a flow classifier',
        [
            {
                'method': 'PUT',
                'path': '/sfc/flow_classifiers/{id}',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'delete_flow_classifier',
        base.RULE_ADMIN_OR_OWNER,
        'Delete a flow classifier',
        [
            {
                'method': 'DELETE',
                'path': '/sfc/flow_classifiers/{id}',
            },
        ]
    ),
    policy.DocumentedRuleDefault(
        'get_flow_classifier',
        base.RULE_ADMIN_OR_OWNER,
        'Get flow classifiers',
        [
            {
                'method': 'GET',
                'path': '/sfc/flow_classifiers',
            },
            {
                'method': 'GET',
                'path': '/sfc/flow_classifiers/{id}',
            },
        ]
    ),
]


def list_rules():
    return rules



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from neutron.agent import rpc as agent_rpc
from neutron import manager
from neutron_lib.agent import l2_extension
from neutron_lib.agent import topics
from neutron_lib import rpc as n_rpc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging

from networking_sfc.services.sfc.drivers.ovs import rpc_topics as sfc_topics

LOG = logging.getLogger(__name__)


class SfcPluginApi():
    def __init__(self, topic, host):
        self.host = host
        self.target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(self.target)

    def update_flowrules_status(self, context, flowrules_status):
        cctxt = self.client.prepare()
        return cctxt.call(
            context, 'update_flowrules_status',
            flowrules_status=flowrules_status)

    def get_flowrules_by_host_portid(self, context, port_id):
        cctxt = self.client.prepare()
        return cctxt.call(
            context, 'get_flowrules_by_host_portid',
            host=self.host, port_id=port_id)


class SfcAgentDriver(metaclass=abc.ABCMeta):
    """Defines stable abstract interface for SFC Agent Driver."""

    @abc.abstractmethod
    def initialize(self):
        """Perform SFC agent driver initialization."""

    def consume_api(self, agent_api):
        """Consume the AgentAPI instance from the SfcAgentExtension class

        :param agent_api: An instance of an agent specific API
        """

    def update_flow_rules(self, flowrule, flowrule_status):
        """Update a flow rule in driver."""

    def delete_flow_rule(self, flowrule, flowrule_status):
        """Delete a flow rule in driver."""


class SfcAgentExtension(l2_extension.L2AgentExtension):

    def initialize(self, connection, driver_type):
        """Initialize agent extension."""
        self.sfc_driver = manager.NeutronManager.load_class_for_provider(
            'networking_sfc.sfc.agent_drivers', driver_type)()
        self.sfc_driver.consume_api(self.agent_api)
        self.sfc_driver.initialize()

        self._sfc_setup_rpc()

    def consume_api(self, agent_api):
        """Receive neutron agent API object

        Allows an extension to gain access to resources internal to the
        neutron agent and otherwise unavailable to the extension.
        """
        self.agent_api = agent_api

    def handle_port(self, context, port):
        """Handle agent SFC extension port add/update."""
        port_id = port['port_id']
        resync = False
        flowrule_status = []
        try:
            LOG.debug("a new device %s is found", port_id)
            flows_list = (
                self.sfc_plugin_rpc.get_flowrules_by_host_portid(
                    context, port_id
                )
            )
            if flows_list:
                for flow in flows_list:
                    self.sfc_driver.update_flow_rules(
                        flow, flowrule_status)
        except Exception as e:
            LOG.exception(e)
            LOG.error("SFC L2 extension handle_port failed")
            resync = True

        if flowrule_status:
            self.sfc_plugin_rpc.update_flowrules_status(
                context, flowrule_status)

        return resync

    def delete_port(self, context, port):
        """Handle agent SFC extension port delete."""
        port_id = port['port_id']
        resync = False
        LOG.info("a device %s is removed", port_id)
        try:
            self._delete_ports_flowrules_by_id(context, port_id)
        except Exception as e:
            LOG.exception(e)
            LOG.error(
                "delete port flow rule failed for %(port_id)s",
                {'port_id': port_id}
            )
            resync = True

        return resync

    def update_flow_rules(self, context, **kwargs):
        flowrule_status = []
        try:
            flowrules = kwargs['flowrule_entries']
            LOG.debug("update_flow_rules received,  flowrules = %s",
                      flowrules)

            if flowrules:
                self.sfc_driver.update_flow_rules(
                    flowrules, flowrule_status)
        except Exception as e:
            LOG.exception(e)
            LOG.error("update_flow_rules failed")

        if flowrule_status:
            self.sfc_plugin_rpc.update_flowrules_status(
                context, flowrule_status)

    def delete_flow_rules(self, context, **kwargs):
        flowrule_status = []
        try:
            flowrules = kwargs['flowrule_entries']
            LOG.debug("delete_flow_rules received,  flowrules= %s", flowrules)
            if flowrules:
                self.sfc_driver.delete_flow_rule(
                    flowrules, flowrule_status)
        except Exception as e:
            LOG.exception(e)
            LOG.error("delete_flow_rules failed")

        if flowrule_status:
            self.sfc_plugin_rpc.update_flowrules_status(
                context, flowrule_status)

    def _sfc_setup_rpc(self):
        self.sfc_plugin_rpc = SfcPluginApi(
            sfc_topics.SFC_PLUGIN, cfg.CONF.host)

        self.topic = sfc_topics.SFC_AGENT
        self.endpoints = [self]
        consumers = [
            [sfc_topics.PORTFLOW, topics.UPDATE],
            [sfc_topics.PORTFLOW, topics.DELETE]
        ]

        # subscribe sfc plugin message
        self.connection = agent_rpc.create_consumers(
            self.endpoints,
            self.topic,
            consumers)

    def _delete_ports_flowrules_by_id(self, context, ports_id):
        flowrule_status = []
        try:
            LOG.debug("delete_port_id_flows received, ports_id= %s", ports_id)
            count = 0
            if ports_id:
                for port_id in ports_id:
                    flowrule = (
                        self.sfc_plugin_rpc.get_flowrules_by_host_portid(
                            context, port_id
                        )
                    )
                    if flowrule:
                        self.sfc.driver.delete_flow_rule(
                            flowrule, flowrule_status)
            LOG.debug(
                "_delete_ports_flowrules_by_id received, count= %s", count)
        except Exception as e:
            LOG.exception(e)
            LOG.error("delete_port_id_flows failed")
        if flowrule_status:
            self.sfc_plugin_rpc.update_flowrules_status(
                context, flowrule_status)



# Copyright 2015 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory

from networking_sfc.db import flowclassifier_db as fc_db
from networking_sfc.extensions import flowclassifier as fc_ext
from networking_sfc.services.flowclassifier.common import context as fc_ctx
from networking_sfc.services.flowclassifier.common import exceptions as fc_exc
from networking_sfc.services.flowclassifier import driver_manager as fc_driver


LOG = logging.getLogger(__name__)


class FlowClassifierPlugin(fc_db.FlowClassifierDbPlugin):

    """Implementation of the Plugin."""
    supported_extension_aliases = [fc_ext.FLOW_CLASSIFIER_EXT]
    path_prefix = fc_ext.FLOW_CLASSIFIER_PREFIX

    def __init__(self):
        self.driver_manager = fc_driver.FlowClassifierDriverManager()
        super(FlowClassifierPlugin, self).__init__()
        self.driver_manager.initialize()

    def _get_port(self, context, id):
        port = super(FlowClassifierPlugin, self)._get_port(context, id)
        return directory.get_plugin().get_port(context, port['id'])

    @log_helpers.log_method_call
    def create_flow_classifier(self, context, flow_classifier):
        with db_api.CONTEXT_WRITER.using(context):
            fc_db = super(FlowClassifierPlugin, self).create_flow_classifier(
                context, flow_classifier)
            fc_db_context = fc_ctx.FlowClassifierContext(self, context, fc_db)
            self.driver_manager.create_flow_classifier_precommit(
                fc_db_context)

        try:
            self.driver_manager.create_flow_classifier_postcommit(
                fc_db_context)
        except fc_exc.FlowClassifierDriverError as e:
            LOG.exception(e)
            with excutils.save_and_reraise_exception():
                LOG.error("Create flow classifier failed, "
                          "deleting flow_classifier '%s'",
                          fc_db['id'])
                self.delete_flow_classifier(context, fc_db['id'])
        return fc_db

    @log_helpers.log_method_call
    def update_flow_classifier(self, context, id, flow_classifier):
        with db_api.CONTEXT_WRITER.using(context):
            original_flowclassifier = self.get_flow_classifier(context, id)
            updated_fc = super(
                FlowClassifierPlugin, self
            ).update_flow_classifier(
                context, id, flow_classifier)
            fc_db_context = fc_ctx.FlowClassifierContext(
                self, context, updated_fc,
                original_flowclassifier=original_flowclassifier)
            self.driver_manager.update_flow_classifier_precommit(fc_db_context)
        try:
            self.driver_manager.update_flow_classifier_postcommit(
                fc_db_context)
        except fc_exc.FlowClassifierDriverError as e:
            LOG.exception(e)
            with excutils.save_and_reraise_exception():
                LOG.error("Update flow classifier failed, "
                          "flow_classifier '%s'",
                          updated_fc['id'])

        return updated_fc

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context, fc_id):
        fc = self.get_flow_classifier(context, fc_id)
        fc_context = fc_ctx.FlowClassifierContext(self, context, fc)
        try:
            self.driver_manager.delete_flow_classifier(fc_context)
        except fc_exc.FlowClassifierDriverError as e:
            LOG.exception(e)
            with excutils.save_and_reraise_exception():
                LOG.error("Delete flow classifier failed, "
                          "flow_classifier '%s'",
                          fc_id)

        with db_api.CONTEXT_WRITER.using(context):
            fc = self.get_flow_classifier(context, fc_id)
            fc_context = fc_ctx.FlowClassifierContext(self, context, fc)
            super(FlowClassifierPlugin, self).delete_flow_classifier(
                context, fc_id)
            self.driver_manager.delete_flow_classifier_precommit(fc_context)
        self.driver_manager.delete_flow_classifier_postcommit(fc_context)
